# Copyright 2022 Tecton, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pandas
from typing import Any, Callable, Collection, Mapping, Sequence, Union

from airflow.models import BaseOperator
from airflow.utils.context import Context

from airflow_tecton.hooks.tecton_hook import TectonHook
from airflow_tecton.operators.df_utils import ingest_feature_table_with_pandas_df
from airflow_tecton.operators.job_utils import wait_until_completion, kill_job


class TectonFeatureTableJobOperator(BaseOperator):
    """
    An Airflow operator that launches a Tecton Feature Table ingestion job,
    and waits for its completion.

    Use this if you want to submit a Tecton job via Airflow and control
    retries via Airflow. Each attempt of this operator creates 1 job.
    """

    template_fields: Sequence[str] = ("templates_dict", "op_args", "op_kwargs")
    template_fields_renderers = {"templates_dict": "json", "op_args": "py", "op_kwargs": "py"}

    def __init__(
        self,
        *,
        conn_id: str = "tecton_default",
        workspace: str,
        feature_view: str,
        online: bool,
        offline: bool,
        df_generator: Callable[..., pandas.DataFrame] = None,
        op_args: Union[Collection[Any], None] = None,
        op_kwargs: Union[Mapping[str, Any], None] = None,
        templates_dict: Union[dict, None] = None,
        templates_exts: Union[list[str], None] = None,
        **kwargs,
    ):
        """

        :param conn_id: Airflow connection ID for Tecton connection
        :param workspace: Workspace of FeatureView
        :param feature_view: FeatureView name
        :param online: Whether job writes to online store
        :param offline: Whether job writes to offline store
        :param df_generator: A reference to an object that is callable and
               returns pandas.DataFrame
        :param op_args: a list of positional arguments that will get unpacked when
            calling df_generator
        :param op_kwargs: a dictionary of keyword arguments that will get unpacked
            in df_generator
        :param templates_dict: a dictionary where the values are templates that
            will get templated by the Airflow engine sometime between
            ``__init__`` and ``execute`` takes place and are made available
            in df generator's context after the template has been applied. (templated)
        :param templates_exts: a list of file extensions to resolve while
            processing templated fields, for examples ``['.sql', '.hql']``
        :param kwargs: Airflow base kwargs passed to BaseOperator
        """
        super().__init__(**kwargs)
        self.workspace = workspace
        self.feature_view = feature_view
        self.online = online
        self.offline = offline
        self.conn_id = conn_id
        self.job_id = None
        if df_generator and not callable(df_generator):
            raise Exception("`df_generator` param must be callable")
        self.df_generator = df_generator
        self.op_args = op_args or ()
        self.op_kwargs = op_kwargs or {}
        self.templates_dict = templates_dict
        if templates_exts:
            self.template_ext = templates_exts

    def execute(self, context: Context) -> Any:
        hook = TectonHook.create(self.conn_id)

        resp = ingest_feature_table_with_pandas_df(hook, self.workspace, self.feature_view, context,
                                                   self.df_generator, self.op_args, self.op_kwargs,
                                                   self.templates_dict)

        self.job_id = resp["job"]["id"]
        wait_until_completion(hook, self.workspace, self.feature_view, self.job_id)

    def on_kill(self) -> None:
        kill_job(TectonHook.create(self.conn_id), self.workspace, self.feature_view, self.job_id)



# Copyright 2022 Tecton, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
def get_provider_info():
    return {
        "package-name": "airflow-tecton",
        "name": "Apache Airflow Providers Tecton",
        "description": "Apache Airflow Providers for Tecton.",
        "hook-class-names": ["airflow_tecton.hooks.tecton_hook.TectonHook"],
        "connection-types": [
            {
                "connection-type": "tecton",
                "hook-class-name": "airflow_tecton.hooks.tecton_hook.TectonHook",
            }
        ],
        "versions": ["0.0.1"],  # Required
    }


from airflow_tecton.operators.tecton_job_operator import TectonJobOperator
from airflow_tecton.operators.tecton_trigger_operator import TectonTriggerOperator
from airflow_tecton.sensors.tecton_sensor import TectonSensor



# Copyright 2022 Tecton, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import datetime
import logging
from typing import Sequence, Union, Optional

from airflow.sensors.base import BaseSensorOperator

from airflow_tecton.hooks.tecton_hook import TectonHook


class TectonSensor(BaseSensorOperator):
    """
    Sensor for Tecton FeatureViews or FeatureServices.

    A workspace and exactly one of (FeatureService name, FeatureView name) must be provided.

    This will return "success" when the given FeatureView or FeatureService
    has data materialized from its feature_start_time (for FeatureServices,
    this is the latest for all FeatureViews listed). Specifying `online`
    and `offline` controls which stores to wait for materialization in.
    """

    template_fields: Sequence[str] = ("ready_time",)

    def __init__(
        self,
        *,
        conn_id: str = "tecton_default",
        workspace: str,
        feature_view: Optional[str] = None,
        feature_service: Optional[str] = None,
        online: bool,
        offline: bool,
        ready_time: Union[str, datetime.datetime] = "{{ data_interval_end }}",
        poke_interval=300,
        **kwargs,
    ):
        """

        :param conn_id: Airflow connection ID for Tecton connection
        :param workspace: workspace name
        :param feature_view: [optional] feature view name (mutually exclusive with feature_service)
        :param feature_service: [optional] feature service name (mutually exclusive with feature_view)
        :param online: Whether to wait for data to be materialized in the online store
        :param offline: Whether to wait for data ot be materialized in the offline store.
        :param ready_time: What time to wait for feature data to be ready at. By default, this is the Airflow template variable "data_interval_end"
        :param poke_interval: How often to check Tecton
        :param kwargs: Airflow base kwargs passed to BaseOperator
        """
        super().__init__(**kwargs, mode="reschedule", poke_interval=poke_interval)
        assert (feature_view is None) != (
            feature_service is None
        ), "Exactly one of feature_view or feature_service must be set"
        assert (
            online or offline
        ), "You must wait for either the online store, offline store, or both. Waiting for neither is a no-op."
        self.conn_id = conn_id
        self.workspace = workspace
        self.feature_view = feature_view
        self.feature_service = feature_service
        self.online = online
        self.offline = offline
        self.ready_time = ready_time

    def _ready_time(self):
        if isinstance(self.ready_time, str):
            return datetime.datetime.fromisoformat(self.ready_time)
        else:
            return self.ready_time

    def poke(self, context) -> bool:
        hook = TectonHook(self.conn_id)
        if self.feature_view:
            readiness_resp = hook.get_latest_ready_time(
                self.workspace, feature_view=self.feature_view
            )
        else:
            readiness_resp = hook.get_latest_ready_time(
                self.workspace, feature_service=self.feature_service
            )
        online_ready = self._maybe_check_readiness(
            readiness_resp, "online", self.online
        )
        offline_ready = self._maybe_check_readiness(
            readiness_resp, "offline", self.offline
        )
        return online_ready and offline_ready

    def _maybe_check_readiness(self, resp, store, should_check) -> bool:
        if should_check:
            actual_time = resp[store + "_latest_ready_time"]
            ready = actual_time and actual_time >= self._ready_time()
            if ready:
                logging.info(f"{store.capitalize()} store ready!")
            else:
                logging.info(
                    f"{store.capitalize()} store not ready. Expected: {self._ready_time()} Actual: {actual_time}"
                )
            return ready
        else:
            return True



# Copyright 2022 Tecton, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import io
from typing import Callable, Any, Collection, Mapping, Optional

import pandas as pd
import requests

from airflow.utils.context import context_merge, Context
from airflow.utils.operator_helpers import KeywordParameters

from airflow_tecton.hooks.tecton_hook import TectonHook


def upload_df_pandas(upload_url: str, df: pd.DataFrame):
    out_buffer = io.BytesIO()
    df.to_parquet(out_buffer, index=False)

    # Maximum 1GB per ingestion
    if out_buffer.__sizeof__() > 1_000_000_000:
        raise Exception(f"Pandas dataframe size, {out_buffer.__sizeof__()}, is larger than 1GB")

    r = requests.put(upload_url, data=out_buffer.getvalue())
    if r.status_code != 200:
        raise Exception("Pandas dataframe upload failed with %s" % r.status_code)
    return r.status_code


def ingest_feature_table_with_pandas_df(hook: TectonHook,
                                        workspace: str,
                                        feature_view: str,
                                        context: Context,
                                        df_generator: Callable,
                                        op_args: Optional[Collection[Any]],
                                        op_kwargs: Optional[Mapping[str, Any]],
                                        templates_dict: Optional[dict[str, Any]],
                                        tecton_managed_retries: bool = False):
    context_merge(context, op_kwargs, templates_dict=templates_dict)
    new_op_kwargs = KeywordParameters.determine(df_generator, op_args, context).unpacking()
    op_kwargs.update(new_op_kwargs)

    df_info = hook.get_dataframe_info(feature_view, workspace)

    df_path = df_info["df_path"]
    upload_url = df_info["signed_url_for_df_upload"]

    df = df_generator(*op_args, **op_kwargs)
    upload_df_pandas(upload_url, df)

    return hook.ingest_dataframe(feature_view, df_path, workspace, tecton_managed_retries)



# Copyright 2022 Tecton, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import datetime
import logging
import pandas
from typing import Any, Callable, Collection, List, Mapping, Sequence, Union

from airflow.models import BaseOperator

from airflow_tecton.hooks.tecton_hook import TectonHook
from airflow_tecton.operators.df_utils import ingest_feature_table_with_pandas_df


class TectonFeatureTableTriggerOperator(BaseOperator):
    """
    An Airflow operator that kicks off a Tecton Feature Table ingestion job,
    and does not wait for its completion.

    Note that this will use Tecton managed retries.

    Use this if you have unpredictably arriving data but want Tecton
    to manage retries of jobs.
    """

    template_fields: Sequence[str] = ("templates_dict", "op_args", "op_kwargs")
    template_fields_renderers = {"templates_dict": "json", "op_args": "py", "op_kwargs": "py"}

    def __init__(
        self,
        *,
        conn_id: str = "tecton_default",
        workspace: str,
        feature_view: str,
        online: bool,
        offline: bool,
        df_generator: Callable[..., pandas.DataFrame] = None,
        op_args: Union[Collection[Any], None] = None,
        op_kwargs: Union[Mapping[str, Any], None] = None,
        templates_dict: Union[dict, None] = None,
        templates_exts: Union[list[str], None] = None,
        **kwargs,
    ):
        """

        :param conn_id: Airflow connection ID for Tecton connection
        :param workspace: Workspace of FeatureView
        :param feature_view: FeatureView name
        :param online: Whether job writes to online store
        :param offline: Whether job writes to offline store
        :param df_generator: A reference to an object that is callable and returns pandas.DataFrame
        :param op_args: a list of positional arguments that will get unpacked when
            calling df_generator
        :param op_kwargs: a dictionary of keyword arguments that will get unpacked
            in df_generator
        :param templates_dict: a dictionary where the values are templates that
            will get templated by the Airflow engine sometime between
            ``__init__`` and ``execute`` takes place and are made available
            in df generator's context after the template has been applied. (templated)
        :param templates_exts: a list of file extensions to resolve while
            processing templated fields, for examples ``['.sql', '.hql']``
        :param kwargs: Airflow base kwargs passed to BaseOperator
        """
        super().__init__(**kwargs)
        self.workspace = workspace
        self.feature_view = feature_view
        self.online = online
        self.offline = offline
        self.conn_id = conn_id
        if df_generator and not callable(df_generator):
            raise Exception("`df_generator` param must be callable")
        self.df_generator = df_generator
        self.op_args = op_args or ()
        self.op_kwargs = op_kwargs or {}
        self.templates_dict = templates_dict
        if templates_exts:
            self.template_ext = templates_exts

    def execute(self, context) -> List[str]:
        hook = TectonHook.create(self.conn_id)

        resp = ingest_feature_table_with_pandas_df(hook, self.workspace, self.feature_view, context,
                                                   self.df_generator, self.op_args, self.op_kwargs,
                                                   self.templates_dict, True)

        new_job_id = resp["job"]["id"]
        logging.info(f"Launched job with id {new_job_id}")
        return [new_job_id]



# Copyright 2022 Tecton, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import datetime
import logging
import pprint
import time
from typing import Any, Sequence, Union

from airflow.models import BaseOperator

from airflow_tecton.hooks.tecton_hook import TectonHook
from airflow_tecton.operators.job_utils import wait_until_completion, kill_job


class TectonJobOperator(BaseOperator):
    """
    An Airflow operator that launches a Tecton job, and waits for its completion. If the latest job with the same params
    is running, it will cancel it first. If the latest job with the same params is successful, this will return without
    running a new job.

    Use this if you want to submit a Tecton job via Airflow and control retries via Airflow. Each attempt of this operator creates 1 job.
    """

    template_fields: Sequence[str] = ("start_time", "end_time")

    def __init__(
        self,
        *,
        conn_id: str = "tecton_default",
        workspace: str,
        feature_view: str,
        online: bool,
        offline: bool,
        start_time: Union[str, datetime.datetime] = "{{ data_interval_start }}",
        end_time: Union[str, datetime.datetime] = "{{ data_interval_end }}",
        allow_overwrite: bool = False,
        **kwargs,
    ):
        """

        :param conn_id: Airflow connection ID for Tecton connection
        :param workspace: Workspace of FeatureView
        :param feature_view: FeatureView name
        :param start_time: Start of time range for materialization job
        :param end_time: End of time range for materialization job
        :param online: Whether job writes to online store
        :param offline: Whether job writes to offline store
        :param allow_overwrite: Whether jobs are able to run materialization for periods that previously have materialized data. Note that this can cause inconsistencies if the underlying data has changed.
        :param kwargs: Airflow base kwargs passed to BaseOperator
        """
        super().__init__(**kwargs)
        self.workspace = workspace
        self.feature_view = feature_view
        self.online = online
        self.offline = offline
        self.start_time = start_time
        self.end_time = end_time
        self.allow_overwrite = allow_overwrite
        self.conn_id = conn_id
        self.job_id = None

    def execute(self, context) -> Any:
        hook = TectonHook.create(self.conn_id)

        job = hook.find_materialization_job(
            workspace=self.workspace,
            feature_view=self.feature_view,
            online=self.online,
            offline=self.offline,
            start_time=self.start_time,
            end_time=self.end_time,
        )
        if job:
            logging.info(f"Existing job found: {pprint.pformat(job)}")
            if job["state"].lower().endswith("running"):
                logging.info("Job in running state; cancelling")
                hook.cancel_materialization_job(
                    workspace=self.workspace,
                    feature_view=self.feature_view,
                    job_id=job["id"],
                )
                while (
                    not hook.get_materialization_job(
                        workspace=self.workspace,
                        feature_view=self.feature_view,
                        job_id=job["id"],
                    )["job"]["state"]
                    .lower()
                    .endswith("manually_cancelled")
                ):
                    logging.info(f"waiting for job to enter state manually_cancelled")
                    time.sleep(60)
            elif job["state"].lower().endswith("success"):
                if self.allow_overwrite:
                    logging.info(
                        "Overwriting existing job in success state; (allow_overwrite=True)"
                    )
                else:
                    logging.info("Existing job in success state; exiting")
                    return

        resp = hook.submit_materialization_job(
            workspace=self.workspace,
            feature_view=self.feature_view,
            online=self.online,
            offline=self.offline,
            start_time=self.start_time,
            end_time=self.end_time,
            allow_overwrite=self.allow_overwrite,
            tecton_managed_retries=False,
        )

        self.job_id = resp["job"]["id"]

        wait_until_completion(hook, self.workspace, self.feature_view, self.job_id)

    def on_kill(self) -> None:
        kill_job(TectonHook.create(self.conn_id), self.workspace, self.feature_view, self.job_id)



# Copyright 2022 Tecton, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import datetime
import logging
import pprint
from typing import List
from typing import Sequence
from typing import Union

from airflow.models import BaseOperator

from airflow_tecton.hooks.tecton_hook import TectonHook


class TectonTriggerOperator(BaseOperator):
    """
    An Airflow operator that kicks off a Tecton job, and does not wait
    for its completion. If the latest job with the same params is in
    the running or success state, this operator will do nothing.

    Note that this will use Tecton managed retries.

    Use this if you have unpredictably arriving data but want Tecton to manage retries of jobs.
    """

    template_fields: Sequence[str] = ("start_time", "end_time")

    def __init__(
        self,
        *,
        conn_id: str = "tecton_default",
        workspace: str,
        feature_view: str,
        online: bool,
        offline: bool,
        start_time: Union[str, datetime.datetime] = "{{ data_interval_start }}",
        end_time: Union[str, datetime.datetime] = "{{ data_interval_end }}",
        allow_overwrite: bool = False,
        **kwargs,
    ):
        """

        :param conn_id: Airflow connection ID for Tecton connection
        :param workspace: Workspace of FeatureView
        :param feature_view: FeatureView name
        :param start_time: Start of time range for materialization job
        :param end_time: End of time range for materialization job
        :param online: Whether job writes to online store
        :param offline: Whether job writes to offline store
        :param allow_overwrite: Whether jobs are able to run materialization for periods that previously have materialized data. Note that this can cause inconsistencies if the underlying data has changed.
        :param kwargs: Airflow base kwargs passed to BaseOperator
        """
        super().__init__(**kwargs)
        self.workspace = workspace
        self.feature_view = feature_view
        self.online = online
        self.offline = offline
        self.start_time = start_time
        self.end_time = end_time
        self.allow_overwrite = allow_overwrite
        self.conn_id = conn_id

    def execute(self, context) -> List[str]:
        hook = TectonHook.create(self.conn_id)

        job = hook.find_materialization_job(
            workspace=self.workspace,
            feature_view=self.feature_view,
            online=self.online,
            offline=self.offline,
            start_time=self.start_time,
            end_time=self.end_time,
        )
        if job:
            logging.info(f"Existing job found: {pprint.pformat(job)}")
            if job["state"].lower().endswith("running"):
                logging.info(f"Job already in running state; not triggering new job")
                return [job["id"]]
            elif job["state"].lower().endswith("success"):
                if self.allow_overwrite:
                    logging.info(
                        "Overwriting existing job in success state; (allow_overwrite=True)"
                    )
                else:
                    logging.info(
                        f"Job already in success state; not triggering new job"
                    )
                    return [job["id"]]
            else:
                logging.info(f"Job in {job['state']} state; triggering new job")

        resp = hook.submit_materialization_job(
            workspace=self.workspace,
            feature_view=self.feature_view,
            online=self.online,
            offline=self.offline,
            start_time=self.start_time,
            end_time=self.end_time,
            allow_overwrite=self.allow_overwrite,
            tecton_managed_retries=True,
        )
        job_id = resp["job"]["id"]
        logging.info(f"Launched job with id {job_id}")
        return [job_id]



# Copyright 2022 Tecton, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="airflow-tecton",
    version="0.1.1",
    description="Provider for using Tecton with Airflow.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    entry_points={
        "apache_airflow_provider": [
            "provider_info=airflow_tecton.__init__:get_provider_info"
        ]
    },
    license="Apache License 2.0",
    packages=[
        "airflow_tecton",
        "airflow_tecton.hooks",
        "airflow_tecton.sensors",
        "airflow_tecton.operators",
    ],
    install_requires=["apache-airflow>=2.0",
                      "requests",
                      "pandas",
                      "pyarrow",
                      "fastparquet"],
    setup_requires=["setuptools", "wheel"],
    test_require=["requests_mock"],
    url="http://tecton.ai/",
    classifiers=[
        "Framework :: Apache Airflow",
        "Framework :: Apache Airflow :: Provider",
    ],
    python_requires=">=3.7",
)



# Copyright 2022 Tecton, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import datetime
import json
import logging
import pprint
import sys
import urllib
from typing import Any
from typing import Dict
from typing import Optional
from typing import Union

import pytz
import requests
from airflow.hooks.base import BaseHook

JOBS_API_BASE = "/api/v1/jobs"
SUBMIT_JOB_METHOD = "submit-materialization-job"
GET_JOB_METHOD = "get-materialization-job"
CANCEL_JOB_METHOD = "cancel-materialization-job"
LIST_JOB_METHOD = "list-materialization-jobs"
READINESS_METHOD = "get-latest-ready-time"
INGEST_DATAFRAME = "ingest-dataframe"
GET_DATAFRAME_INFO = "get-dataframe-info"


class TectonHook(BaseHook):
    """
    Interact with Tecton's API. Currently supported are:
        get_materialization_job
        list_materialization_jobs
        submit_materialization_job
        cancel_materialization_job
        get_latest_ready_time

    To add Connection:
        Put `https://your-tecton-url.tecton.ai` as host in the Connections page
        Put your Tecton API key in the password field
        The default connection is `tecton_default`
    """

    conn_name_attr = "tecton_conn_id"
    default_conn_name = "tecton_default"
    conn_type = "tecton"
    hook_name = "Tecton"

    def __init__(self, conn_id: str = default_conn_name):
        super().__init__()
        self.conn = self.get_connection(conn_id)
        self._session = None

    def _create_conn(self):
        session = requests.Session()

        conn = self.conn

        if conn.host and conn.host.startswith("http"):
            self.base_url = conn.host
        else:
            host = "https://" + conn.host
            self.base_url = host

        session.headers.update({"Authorization": f"Tecton-key {conn.password}"})
        session.headers.update({"Content-type": "application/json"})
        self._session = session

    def get_conn(self) -> requests.Session:
        if self._session is None:
            self._create_conn()
        return self._session

    def _pformat_dict(self, d):
        if sys.version_info >= (3, 8):
            # this flag does not exist in earlier python versions
            return pprint.pformat(d, sort_dicts=False)
        else:
            return pprint.pformat(d)

    def _make_request(
        self,
        conn: requests.Session,
        url: str,
        data: Dict[str, Any],
        verbose: bool = False,
    ) -> Dict[str, Any]:
        full_path = urllib.parse.urljoin(self.base_url, url)
        num_retries = 3

        for i in range(num_retries):
            if verbose:
                logging.info(
                    f"Making POST request to {full_path} with body=\n{self._pformat_dict(data)}"
                )

            try:
                resp = conn.post(full_path, json.dumps(data))
            except (
                requests.exceptions.ConnectionError,
                requests.exceptions.Timeout,
            ) as re:
                if verbose:
                    logging.info(
                        f"Received an error for attempt {i + 1}/{num_retries}. Error: {re}"
                    )
                    logging.info(f"Retrying request to {full_path}")
                # Refresh session
                self._create_conn()
                continue

            try:
                resp.raise_for_status()
            except Exception as e:
                exc = e
                if "error" in resp.json():
                    raise Exception(f"Tecton error: {resp.json()['error']}")
                raise e
            else:
                exc = None
            finally:
                if verbose or exc:
                    logging.info(
                        f"Response: Code={resp.status_code} Body=\n{self._pformat_dict(resp.json())}"
                    )
            return resp.json()

        raise Exception(f"Max retries reached making request to {full_path}")

    def _canonicalize_datetime(self, dt) -> str:
        if isinstance(dt, str):
            try:
                return datetime.datetime.fromisoformat(dt).strftime(
                    "%Y-%m-%dT%H:%M:%SZ"
                )
            except:
                return self._parse_time(dt).strftime("%Y-%m-%dT%H:%M:%SZ")
        elif isinstance(dt, datetime.datetime):
            return dt.strftime("%Y-%m-%dT%H:%M:%SZ")
        else:
            raise Exception("unexpected type for datetime: " + str(type(dt)))

    def _parse_time(self, dt):
        try:
            return datetime.datetime.strptime(dt, "%Y-%m-%dT%H:%M:%SZ").replace(
                tzinfo=pytz.UTC
            )
        except:
            return datetime.datetime.strptime(dt, "%Y-%m-%dT%H:%M:%S.%fZ").replace(
                tzinfo=pytz.UTC
            )

    def find_materialization_job(
        self,
        workspace: str,
        feature_view: str,
        start_time: Union[datetime.datetime, str],
        end_time: Union[datetime.datetime, str],
        online: bool,
        offline: bool,
        job_type: str = "batch",
    ):
        jobs = [
            x
            for x in self.list_materialization_jobs(
                workspace=workspace, feature_view=feature_view
            ).get("jobs", [])
        ]
        for job in sorted(
            jobs, reverse=True, key=lambda x: self._parse_time(x["created_at"])
        ):
            self.log.info("job = %s" % job)
            if job_type.lower() == job.get("job_type", "").lower():
                if (
                    job["online"] == online
                    and job["offline"] == offline
                    and (job["start_time"] == self._canonicalize_datetime(start_time))
                    and (job["end_time"] == self._canonicalize_datetime(end_time))
                ):
                    return job
        return None

    def submit_materialization_job(
        self,
        workspace: str,
        feature_view: str,
        start_time: Union[datetime.datetime, str],
        end_time: Union[datetime.datetime, str],
        online: bool,
        offline: bool,
        allow_overwrite: bool = False,
        tecton_managed_retries: bool = True,
    ) -> Dict:
        """
        Calls the /jobs/submit-materialization-job API. This submits a Tecton materialization job for a FeatureView.

        :param workspace: Workspace of FeatureView
        :param feature_view: FeatureView name
        :param start_time: Start of time range for materialization job
        :param end_time: End of time range for materialization job
        :param online: Whether job writes to online store
        :param offline: Whether job writes to offline store
        :param allow_overwrite: Whether the job should be able to overwrite an existing, successful job. Note that this can cause inconsistencies if the underlying data has changed.
        :param tecton_managed_retries: Whether the job should be retried by Tecton automatically. Set to `False` if you want to control and submit retries manually.
        :return: The response from the Tecton API. Throws an exception if request was not successful.
        """
        data = {
            "workspace": workspace,
            "feature_view": feature_view,
            "start_time": self._canonicalize_datetime(start_time),
            "end_time": self._canonicalize_datetime(end_time),
            "online": online,
            "offline": offline,
            "overwrite": allow_overwrite,
            "use_tecton_managed_retries": tecton_managed_retries,
        }
        return self._make_request(
            self.get_conn(), f"{JOBS_API_BASE}/{SUBMIT_JOB_METHOD}", data, verbose=True
        )

    def list_materialization_jobs(self, workspace: str, feature_view: str) -> Dict:
        """
        Lists all materialization jobs for a Feature View.
        :param workspace: Workspace name
        :param feature_view: Feature View name
        :return: A dict which contains a list of jobs.
        """
        data = {
            "workspace": workspace,
            "feature_view": feature_view,
        }
        return self._make_request(
            self.get_conn(), f"{JOBS_API_BASE}/{LIST_JOB_METHOD}", data
        )

    def get_materialization_job(
        self, workspace: str, feature_view: str, job_id: str
    ) -> Dict:
        """
        Gets a job
        :param workspace: Workspace name
        :param feature_view: Feature View name
        :param job_id: Job ID
        :return:
        """
        data = {"workspace": workspace, "feature_view": feature_view, "job_id": job_id}
        return self._make_request(
            self.get_conn(), f"{JOBS_API_BASE}/{GET_JOB_METHOD}", data
        )

    def cancel_materialization_job(
        self, workspace: str, feature_view: str, job_id: str
    ) -> Dict:
        """
        Cancels a running materialization job

        :param workspace: Workspace name
        :param feature_view: Feature View name
        :param job_id:  Job ID
        :return:
        """
        data = {"workspace": workspace, "feature_view": feature_view, "job_id": job_id}
        return self._make_request(
            self.get_conn(), f"{JOBS_API_BASE}/{CANCEL_JOB_METHOD}", data, verbose=True
        )

    def get_latest_ready_time(
        self,
        workspace: str,
        feature_view: Optional[str] = None,
        feature_service: Optional[str] = None,
    ) -> Dict:
        """

        :param workspace: workspace name
        :param feature_view: [optional] feature view name: mutually exclusive with feature_service
        :param feature_service: [optional] feature service name: mutually exclusive with feature_view
        :return:
        """
        assert (feature_service is None) != (
            feature_view is None
        ), "Exactly one of feature_view or feature_service should be set"
        data = {
            "workspace": workspace,
        }
        if feature_view:
            data["feature_view"] = feature_view
        if feature_service:
            data["feature_service"] = feature_service
        result = self._make_request(
            self.get_conn(), f"{JOBS_API_BASE}/{READINESS_METHOD}", data
        )
        online_key = "online_latest_ready_time"
        offline_key = "offline_latest_ready_time"
        if online_key not in result:
            result[online_key] = None
        elif result[online_key]:
            result[online_key] = self._parse_time(result[online_key])

        if offline_key not in result:
            result[offline_key] = None
        elif result[offline_key]:
            result[offline_key] = self._parse_time(result[offline_key])

        return result

    def get_dataframe_info(self, feature_view: str, workspace: str):
        """
        Get ingest data frame information

        :param feature_view: feature view name
        :param workspace:  workspace name
        :return:
        """
        data = {"feature_view": feature_view, "workspace": workspace}
        return self._make_request(
            self.get_conn(), f"{JOBS_API_BASE}/{GET_DATAFRAME_INFO}", data, verbose=True
        )

    def ingest_dataframe(
        self,
        feature_view: str,
        df_path: str,
        workspace: str,
        tecton_managed_retries: bool = False,
    ):
        """
        Ingest data frame to FeatureTable from s3 path

        :param feature_view: feature view name
        :param df_path: s3 path containing pandas dataframe data
        :param workspace:  workspace name
        :param tecton_managed_retries: Whether the job should be retried by Tecton automatically.
               Set to `False` if you want to control and submit retries manually.
        :return:
        """
        data = {
            "feature_view": feature_view,
            "df_path": df_path,
            "workspace": workspace,
            "use_tecton_managed_retries": tecton_managed_retries,
        }
        return self._make_request(
            self.get_conn(), f"{JOBS_API_BASE}/{INGEST_DATAFRAME}", data, verbose=True
        )

    @classmethod
    def create(cls, conn_id: str):
        return cls(conn_id)

    @classmethod
    def get_ui_field_behaviour(cls) -> Dict[str, Any]:
        """Returns custom field behaviour"""
        return {
            "hidden_fields": ["port", "schema", "login", "extra"],
            "placeholders": {
                "password": "Output of `tecton api-key create`",
                "host": "https://example.tecton.ai",
            },
            "relabeling": {
                "password": "API Key",
            },
        }



# Copyright 2022 Tecton, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import pprint
import time
from typing import Optional

from airflow_tecton.hooks.tecton_hook import TectonHook


def wait_until_completion(hook: TectonHook,
                          workspace: str,
                          feature_view: str,
                          job_id: str):
    """
    Wait for materialization job to finish successfully.
    If the job fails, throw an exception.
    """
    job_result = hook.get_materialization_job(
        workspace, feature_view, job_id
    )["job"]
    while job_result.get("state", "").upper().endswith("RUNNING"):
        if len(job_result.get("attempts", [])) > 0:
            attempts = job_result["attempts"]
            latest_attempt = attempts[-1]
            logging.info(
                f"Latest attempt #{len(attempts)} in state {latest_attempt['state']} with URL {latest_attempt['run_url']}"
            )
        else:
            logging.info(f"No attempt launched yet")
        time.sleep(60)
        job_result = hook.get_materialization_job(
            workspace, feature_view, job_id
        )["job"]
    if job_result["state"].upper().endswith("SUCCESS"):
        return
    else:
        raise Exception(
            f"Final job state was {job_result['state']}. Final response:\n {job_result}"
        )


def kill_job(hook: TectonHook,
             workspace: str,
             feature_view: str,
             job_id: Optional[str]):
    """
    kill job with job_id when job_id is available.
    """
    if job_id:
        logging.info(f"Attempting to kill job {job_id}")
        hook.cancel_materialization_job(
            workspace, feature_view, job_id
        )
        logging.info(f"Successfully killed job {job_id}")
    else:
        logging.debug(f"No job started; none to kill")
